package com.cmm.common.text;
import com.cmm.common.app.CompositionFactory;
import com.cmm.common.display.MovieClip;
import com.cmm.common.display.SpriteContainer;
import flambe.Component;
import flambe.display.FillSprite;
import flambe.display.Sprite;
import flambe.input.PointerEvent;
import flambe.math.Matrix;
import flambe.math.Point;
import flambe.math.Rectangle;
import flambe.System;
import com.cmm.common.app.Wrapper;

#if html
import js.Browser;
import js.html.TextAreaElement;
#end


/**
 * ...
 * @author Joshua Thom
 */
class InputTextSprite extends SpriteContainer
{
	//PUBLIC PROPERTIES
	public var fontSize(get_fontSize, set_fontSize):Int;
	public var fontFamily(get_fontFamily, set_fontFamily):String;
	public var textAlign(get_textAlign, set_textAlign):String;
	public var color(get_color, set_color):Int;
	
	//public var x(get_x, set_x):Float;
	//public var y(get_y, set_y):Float;
	public var width(get_width, set_width):Float;
	//public var height(get_height, set_height):Float;
	
	//public var visible(get_visible, set_visible):Bool;
	public var maxlength(get_maxlength, set_maxlength):Int;
	public var type(get_type, set_type):String;
	
	public var outline(get_outline, set_outline):Bool;
	public var border(get_border, set_border):Bool;
	public var background(get_background, set_background):Bool;
	
	public var text(get_text, set_text):String;
	
	@:keep public function new(id:String = null) 
	{
		super();
		
		System.pointer.up.connect(function (e) onClick(e));
		System.stage.resize.connect(function () onResize());
		
		if (id != null)
		{
			_format = DynamicTextLibrary.getFormat_input(id);
		}
		else 
		{
			_format = 
			{
				"id" : null,
				"style" : 
				{
					"fontFamily" : "Times New Roman",
					"fontSize" : "12px",
					"textAlign" : "left",
					"fontStyle" : "normal",
					"fontVariant" : "normal",
					"fontWeight" : "normal",
					"color" : "#000000",
					"maxlength" : -1,
					"type" : "text",
					"outline" : false,
					"border" : false,
					"background" : false
				},
				"textMetrics" : 
				{
					"top" : 0,
					"left" : 0,
					"width" : 100,
					"height" : 100
				},
				"text" : "ENTER TEXT"
			}
		}
		
		if (DynamicTextLibrary.settings.useDeviceFonts)
		{
			this.fontFamily = "sans-serif";
		}
		
		this.visible = false;
		
		#if html
		var content:Dynamic = Browser.document.getElementById("content");
		//var form:Dynamic = Browser.document.createElement("form");
		_input = Browser.document.createTextAreaElement();
		//content.appendChild(form);
		//form.appendChild(_input);
		content.appendChild(_input);
		_input.style.position = "absolute";
		_input.style.resize = "none";
		//_input.rows = "4";
		_input.style.width = Std.string(_format.textMetrics.width)+"px";
		_input.style.height = Std.string(_format.textMetrics.height)+"px";
		//_input.wrap = "hard";
		//_input.onkeydown = onKeyDown;
		//_input.onkeypress = onKeyDown;
		_input.onkeyup = onKeyUp;
		_input.onfocus = onFocus;
		_input.onblur = onBlur;
		_input.placeholder = _format.text;
		_placeholder = _input.placeholder;
		_format.text = "";
		#end
		
		updateProps();
		
		//TODO: consider adding this compose to sprite container
		if (id != null)
		{
			compose();
		}
		
	}
	
	private function compose() 
	{
		//trace(this.symbol.name + "COMPOSE");
		if (_components == null)
		{
			_components = [];
		}
			
		
		var compositionObject:CompositionObject = CompositionFactory.getCompositionObjectByClassName(_format.id);
		
		if (compositionObject != null)
		{
			for (componentClass in compositionObject.components) 
			{
				_components.push(Type.createInstance(componentClass, [_format.id, compositionObject.params]));
			}
		}
	}
	
	#if html
	private function onKeyUp(e:Dynamic):Void 
	{
		if (_input != null)
		{
			_format.text = _input.value;
		}
	}
	private function onFocus(e:Dynamic):Void 
	{
		if (_input != null)
		{
			_input.placeholder = '';
		}
	}
	private function onBlur(e:Dynamic):Void 
	{
		if (_input != null)
		{
			_input.placeholder = _placeholder;
		}
	}
	#end
	
	override public function onAddedToStage() 
	{
		super.onAddedToStage();
		if (_components != null)
		{
			for (component in _components) 
			{
				owner.add(component);
			}
		}
		
		this.visible = true;
		
		if (this.numChildren == 0)
		{
			var child:FillSprite = new FillSprite(0xff0000, 10, 10);
			this.addChild(child);
			child.visible = false;
		}
		
		//It is necessary to listen for all parent's changes since the text fields isn't actually tied to sprite.
		//This is a subsitute for doing the translation calculations in onUpdate
		this.x.watch(onXChange);
		this.y.watch(onYChange);
		var parent:SpriteContainer = this.parent;
		while (parent != null)
		{
			parent.x.watch(onXChange);
			parent.y.watch(onYChange);
			parent = parent.parent;
		}
	}
	
	override public function onRemovedFromStage():Void 
	{
		super.onRemovedFromStage();
		this.visible = false;
		
		trace("onRemovedFromStage");
	}
	
	//TODO: Move to a global area
	private function onClick(e:PointerEvent):Void 
	{
		if (_input != null)
		{
			#if html
			var document:Dynamic = Browser.document;
			document.activeElement.blur();
			#end
		}
	}
	
	private function onResize():Void 
	{
		#if html
		var document:Dynamic = Browser.document;
		document.activeElement.blur();
		#end
		
		var scaledX:Int = Std.int(Main.GLOBAL_SCALE * this.x._ + _format.textMetrics.left + Wrapper.instance.gameLayer.x._);
		_input.style.left = Std.string(scaledX) + "px";
		var scaledY:Int = Std.int(Main.GLOBAL_SCALE * this.y._ + _format.textMetrics.top + Wrapper.instance.gameLayer.y._);
		_input.style.top = Std.string(scaledY) + "px";
		
		updateProps();
	}
	
	//FONT FAMILY
	private function get_fontFamily():String {return _format.style.fontFamily;}
	private function set_fontFamily(fontFamily:String):String 
	{
		_format.style.fontFamily = fontFamily;
		updateProps();
		return _format.style.fontFamily;
	}
	
	//FONT SIZE
	private function get_fontSize():Int {return Std.parseInt(_format.style.fontSize);}
	private function set_fontSize(fontSize:Int):Int 
	{
		_format.style.fontSize = Std.string(fontSize) + "px";
		updateProps();
		return Std.parseInt(_format.style.fontSize);
	}
	
	//TEXT ALIGN
	private function get_textAlign():String {return _format.style.textAlign;}
	private function set_textAlign(textAlign:String):String 
	{
		_format.style.textAlign = textAlign;
		updateProps();
		return _format.style.textAlign;
	}
	
	//COLOR
	private function get_color():Int {return Std.parseInt(_format.style.color);}
	private function set_color(color:Int):Int 
	{
		_format.style.color = "#" + StringTools.hex(color,6);
		updateProps();
		return Std.parseInt(_format.style.color);
	}
	
	//TEXT POSITION
	private function get_x():Float {return _x;}
	private function set_x(x:Float):Float 
	{
		_x = x;
		updateProps();
		return _x;
	}
	
	private function get_y():Float {return _y;}
	private function set_y(y:Float):Float 
	{
		_y = y;
		updateProps();
		return _y;
	}
	
	//TEXT DIMENSIONS
	private function get_width():Float {return _width;}
	private function set_width(width:Float):Float 
	{
		_width = width;
		updateProps();
		return _width;
	}
	
	private function get_height():Float {return _height;}
	private function set_height(height:Float):Float 
	{
		_height = height;
		updateProps();
		return _height;
	}
	
	//VISIBILITY
	//private function get_visible():Bool {return _visible;}
	//private function set_visible(visible:Bool):Bool 
	//{
		//_visible = visible;
		//updateProps();
		//return _visible;
	//}
	//override private function get_visible():Bool 
	//{
		//return super.get_visible();
	//}
	override private function set_visible(visible:Bool):Bool 
	{
		super.set_visible(visible);
		updateProps();
		return visible;
	}
	
	//MAX CHARACTER LENGTH
	private function get_maxlength():Int {return _format.style.maxlength;}
	private function set_maxlength(maxlength:Int):Int 
	{
		_format.style.maxlength = maxlength;
		updateProps();
		return _format.style.maxlength;
	}
	
	//INPUT TEXT TYPE
	private function get_type():String {return _format.style.type;}
	private function set_type(type:String):String 
	{
		_format.style.type = type;
		updateProps();
		return _format.style.type;
	}
	
	//OUTLINE BORDER BACKGROUND
	private function get_outline():Bool {return _format.style.outline;}
	private function set_outline(outline:Bool):Bool 
	{
		_format.style.outline = outline;
		updateProps();
		return _format.style.outline;
	}
	
	private function get_border():Bool {return _format.style.border;}
	private function set_border(border:Bool):Bool 
	{
		_format.style.border = border;
		updateProps();
		return _format.style.border;
	}
	
	private function get_background():Bool {return _format.style.background;}
	private function set_background(background:Bool):Bool 
	{
		_format.style.background = background;
		updateProps();
		return _format.style.background;
	}
	
	//TEXT VALUE
	private function get_text():String 
	{
		return _format.text;
	}
	private function set_text(text:String):String 
	{
		_format.text = text;
		updateProps();
		return _format.text;
	}
	
	//Keep track of position updates
	private function onXChange(newValue:Float, oldValue:Float):Void 
	{
		if (_input != null)
		{
			var bounds:Rectangle = Sprite.getBounds(this.owner);
			var matrix:Matrix = this.getViewMatrix();
			var globalCoord:Point = localToGlobal();
			var globalBounds:Rectangle = new Rectangle(globalCoord.x-(matrix.m02-bounds.x), globalCoord.y-(matrix.m12-bounds.y), bounds.width, bounds.height);
			
			//var bounds:Rectangle = Sprite.getBounds(this.owner);
			var scaledX:Int = Std.int(Main.GLOBAL_SCALE * globalBounds.x + _format.textMetrics.left + Wrapper.instance.gameLayer.x._);
			_input.style.left = Std.string(scaledX) + "px";
		}
	}
	
	private function onYChange(newValue:Float, oldValue:Float):Void 
	{
		if (_input != null)
		{
			var bounds:Rectangle = Sprite.getBounds(this.owner);
			var matrix:Matrix = this.getViewMatrix();
			var globalCoord:Point = localToGlobal();
			var globalBounds:Rectangle = new Rectangle(globalCoord.x-(matrix.m02-bounds.x), globalCoord.y-(matrix.m12-bounds.y), bounds.width, bounds.height);
			
			//var bounds:Rectangle = Sprite.getBounds(this.owner);
			var scaledY:Int = Std.int(Main.GLOBAL_SCALE * globalBounds.y + _format.textMetrics.top + Wrapper.instance.gameLayer.y._);
			_input.style.top = Std.string(scaledY) + "px";
		}
	}
	
	override public function onUpdate(dt:Float):Void 
	{
		super.onUpdate(dt);
		//This hack is necessary when an input field is inside a movieclip on the timeline
		if (_initialized == false && this.parent != null && Std.is(this.parent, MovieClip))
		{
			var bounds:Rectangle = Sprite.getBounds(this.owner);
			//TODO: this trick won't work when an input text field is put at 0,0 deliberately...
			if (bounds.x != 0 && bounds.y != 0)
			{
				onXChange(0, 0);
				onYChange(0, 0);
				_initialized = true;
			}
		}
	}
	
	//updates the input text field based on all the custom properties
	private function updateProps():Void 
	{
		//TODO: update the text box with the latest properties
		if (_input != null)
		{
			_input.style.fontFamily = _format.style.fontFamily;
			var scaledFontSize:Int = Std.int(Main.GLOBAL_SCALE * this.fontSize);
			_input.style.fontSize = Std.string(scaledFontSize) + "px";
			_input.style.textAlign = _format.style.textAlign;
			_input.style.color = _format.style.color;
			
			//var scaledX:Int = Std.int(Main.GLOBAL_SCALE * this.x._ + _format.textMetrics.left + Wrapper.instance.gameLayer.x._);
			//var scaledY:Int = Std.int(Main.GLOBAL_SCALE * this.y._ + _format.textMetrics.top + Wrapper.instance.gameLayer.y._);
			//_input.style.left = Std.string(scaledX) + "px";
			//_input.style.top = Std.string(scaledY) + "px";
			
			if (this.parent != null && Std.is(this.parent, MovieClip))
			{
				var mc:MovieClip = cast this.parent;
			
				
				//if (this.numChildren == 0)
				//{
					//this.addChild(new FillSprite(0xff0000, 1, 1));
				//}
				
				var ltg:Point = this.localToGlobal();
				//trace("POS " + ltg);
				
				var bounds:Rectangle = Sprite.getBounds(this.owner);
				var matrix:Matrix = this.getViewMatrix();
				//TODO: don't know if this will work for all cases -- this is only tested for buttons that are centered and should be relatively the same position in global vs local
				var globalCoord:Point = localToGlobal();
				var globalBounds:Rectangle = new Rectangle(globalCoord.x-(matrix.m02-bounds.x), globalCoord.y-(matrix.m12-bounds.y), bounds.width, bounds.height);
				
				trace("POS " + globalCoord + " " + bounds + " " + globalBounds + " " + this.numChildren + " " + this.getViewMatrix().m02);
				
				onXChange(0, 0);
				onYChange(0, 0);
			}
			
			
			
			var scaledWidth:Int = Std.int(Main.GLOBAL_SCALE * _format.textMetrics.width);
			var scaledHeight:Int = Std.int(Main.GLOBAL_SCALE * _format.textMetrics.height);
			_input.style.width = Std.string(scaledWidth) + "px";
			//_input.style.height = Std.string(scaledHeight) + "px";
			
			visible ? _input.style.display = "inline-block" : _input.style.display = "none";
			_input.setAttribute("maxlength", Std.string(_format.style.maxlength));
			//_input.type = "text";// _format.style.type;
			
			_format.style.outline ? _input.style.outline = "solid" : _input.style.outline = "none";
			_format.style.border ? _input.style.border = "2px inset" : _input.style.border = "none";
			_format.style.background ? _input.style.background = "white" : _input.style.background = "none";
			
			_input.value = _format.text;

		}
	}
	
	//PRIVATE VARIABLES
	#if html
	private var _input:TextAreaElement;
	#elseif flash
	private var _input:Dynamic;
	#end
	private var _format:DynamicTextFormat.InputFormat;
	private var _components:Array<Component>;
	
	//private var _fontFamily:String = "Times New Roman";
	//private var _fontSize:Int = 12;
	//private var _textAlign:String = "left";
	//private var _color:String = "#000000";
	
	private var _x:Float = 0;
	private var _y:Float = 0;
	private var _width:Float = 100;
	private var _height:Float = 100;
	
	private var _placeholder:String = "";

	//private var _visible:Bool = true;	
	//private var _maxlength:Int = -1;
	//private var _type:String = InputTextType.TEXT;
		
	//TODO: make these more customizable
	//private var _outline:Bool = false;
	//private var _border:Bool = false;
	//private var _background:Bool = false;
	
	//private var _text:String = "SAMPLE TEXT";
	
	private var _initialized:Bool = false;
}

@final class InputTextType
{
	inline public static var TEXT:String = "text";
	inline public static var EMAIL:String = "email";
	inline public static var PASSWORD:String = "password";
	inline public static var DATE:String = "date";
	inline public static var TELEPHONE:String = "tel";
	inline public static var COLOR:String = "color";
	inline public static var RANGE:String = "range";
	//all possible html <input> types 10/9/2013
	//TODO: not all of these will work in this context
	/*
	button			Defines a clickable button (mostly used with a JavaScript to activate a script)
	checkbox		Defines a checkbox
	color			Defines a color picker
	date			Defines a date control (year, month and day (no time))
	datetime		Defines a date and time control (year, month, day, hour, minute, second, and fraction of a second, based on UTC time zone)
	datetime-local	Defines a date and time control (year, month, day, hour, minute, second, and fraction of a second (no time zone)
	email			Defines a field for an e-mail address
	file			Defines a file-select field and a "Browse..." button (for file uploads)
	hidden			Defines a hidden input field
	image			Defines an image as the submit button
	month			Defines a month and year control (no time zone)
	number			Defines a field for entering a number
	password		Defines a password field (characters are masked)
	radio			Defines a radio button
	range			Defines a control for entering a number whose exact value is not important (like a slider control)
	reset			Defines a reset button (resets all form values to default values)
	search			Defines a text field for entering a search string
	submit			Defines a submit button
	tel				Defines a field for entering a telephone number
	text			Default. Defines a single-line text field (default width is 20 characters)
	time			Defines a control for entering a time (no time zone)
	url				Defines a field for entering a URL
	week			Defines a week and year control (no time zone)
	*/
}





